"""
Supported units:
    - simple RNN ("rnn")
    - standard LSTM ("lstm")
    - SC-LSTM-I ("lstm_skip"), SC-LSTM-P ("lstm_pskip")
    - RRN ("residual")
"""
__author__ = 'v-yirwan'

import theano.tensor as tensor
from Util import *

layers = {'rnn': ('param_init_rnn', 'rnn_layer'),
          'lstm': ('param_init_lstm', 'lstm_layer'),
          'lstm_skip': ('param_init_lstm', 'lstm_skip_layer'),
          'lstm_pskip': ('param_init_lstm_pskip', 'lstm_pskip_layer'),
          'residual': ('param_init_residual', 'residual_layer'),
          }

def _p(pp, name):
    return '%s_%s' % (pp, name)

def get_layer(name):
    fns = layers[name]
    return (eval(fns[0]), eval(fns[1]))

# ===========================
# LSTM-related units
#   - LSTM
#   - SC-LSTM-I, SC-LSTM-P
# ===========================

def param_init_lstm(options, params, prefix='lstm', hier_level=False):
    """
    Initialize the LSTM parameter
    Support hierarchical architecture
    """
    if hier_level:
        # bug fixed: dimension matching for hier-mode
        W = numpy.concatenate([ortho_weight(options['dim_proj']),
                               ortho_weight(options['dim_proj']),
                               ortho_weight(options['dim_proj']),
                               ortho_weight(options['dim_proj'])], axis=1)
    else:
        # bug fixed: different dim for embedding and hidden state
        W = numpy.concatenate([norm_weight(options['dim_word'], options['dim_proj']),
                               norm_weight(options['dim_word'], options['dim_proj']),
                               norm_weight(options['dim_word'], options['dim_proj']),
                               norm_weight(options['dim_word'], options['dim_proj'])], axis=1)
    params[_p(prefix, 'W')] = W
    U = numpy.concatenate([ortho_weight(options['dim_proj']),
                           ortho_weight(options['dim_proj']),
                           ortho_weight(options['dim_proj']),
                           ortho_weight(options['dim_proj'])], axis=1)
    params[_p(prefix, 'U')] = U
    b = numpy.zeros((4 * options['dim_proj'],))
    params[_p(prefix, 'b')] = b.astype(config.floatX)

    return params

def lstm_layer(tparams, state_below, options, prefix='lstm', mask=None):
    nsteps = state_below.shape[0]
    if state_below.ndim == 3:
        n_samples = state_below.shape[1]
    else:
        n_samples = 1

    assert mask is not None

    def _slice(_x, n, dim):
        if _x.ndim == 3:
            return _x[:, :, n * dim:(n + 1) * dim]
        return _x[:, n * dim:(n + 1) * dim]

    def _step(m_, x_, h_, c_):
        preact = tensor.dot(h_, tparams[_p(prefix, 'U')])
        preact += x_

        i = tensor.nnet.sigmoid(_slice(preact, 0, options['dim_proj']))
        f = tensor.nnet.sigmoid(_slice(preact, 1, options['dim_proj']))
        o = tensor.nnet.sigmoid(_slice(preact, 2, options['dim_proj']))
        c = tensor.tanh(_slice(preact, 3, options['dim_proj']))

        c = f * c_ + i * c
        c = m_[:, None] * c + (1. - m_)[:, None] * c_

        h = o * tensor.tanh(c)
        h = m_[:, None] * h + (1. - m_)[:, None] * h_

        return h, c

    state_below = (tensor.dot(state_below, tparams[_p(prefix, 'W')]) +
                   tparams[_p(prefix, 'b')])

    dim_proj = options['dim_proj']
    rval, updates = theano.scan(_step,
                                sequences=[mask, state_below],
                                outputs_info=[tensor.alloc(numpy_floatX(0.),
                                                           n_samples,
                                                           dim_proj),
                                              tensor.alloc(numpy_floatX(0.),
                                                           n_samples,
                                                           dim_proj)],
                                name=_p(prefix, '_layers'),
                                n_steps=nsteps,
                                truncate_gradient=options['truncate_grad'])
    return rval[0]

def lstm_skip_layer(tparams, state_below, options, prefix='lstm_skip', mask=None):
    '''
    lstm layer with soft identity skip connections
    '''
    nsteps = state_below.shape[0]
    n_skip = options['skip_steps']
    if state_below.ndim == 3:
        n_samples = state_below.shape[1]
    else:
        n_samples = 1

    assert mask is not None

    def _slice(_x, n, dim):
        if _x.ndim == 3:
            return _x[:, :, n * dim:(n + 1) * dim]
        return _x[:, n * dim:(n + 1) * dim]

    def _lstm_unit(m_, x_, h_, c_, h_skip, hcnt):
        skip_flag = tensor.eq(hcnt % n_skip, 0)
        preact = tensor.dot(h_, tparams[_p(prefix, 'U')])
        preact += x_

        # gates
        i = tensor.nnet.sigmoid(_slice(preact, 0, options['dim_proj']))
        f = tensor.nnet.sigmoid(_slice(preact, 1, options['dim_proj']))
        o = tensor.nnet.sigmoid(_slice(preact, 2, options['dim_proj']))
        c = tensor.tanh(_slice(preact, 3, options['dim_proj']))

        # cell state
        c = f * c_ + i * c
        c = m_[:, None] * c + (1. - m_)[:, None] * c_
        # new hidden stae
        h = o * tensor.tanh(c) + h_skip * skip_flag
        h = m_[:, None] * h + (1. - m_)[:, None] * h_
        # update h_skip
        h_skip = h_skip * (1-skip_flag) + h * skip_flag
        hcnt += 1

        return h, c, h_skip, hcnt

    state_below = (tensor.dot(state_below, tparams[_p(prefix, 'W')]) +
                   tparams[_p(prefix, 'b')])

    dim_proj = options['dim_proj']
    h = tensor.alloc(numpy_floatX(0.), n_samples, dim_proj)
    c = tensor.alloc(numpy_floatX(0.), n_samples, dim_proj)
    h_skip = tensor.alloc(numpy_floatX(0.), n_samples, dim_proj)
    hcnt = tensor.zeros_like(theano.shared(10.).astype('float32'))
    rval, updates = theano.scan(_lstm_unit,
                                sequences=[mask, state_below],
                                outputs_info=[h, c, h_skip, hcnt],
                                name=_p(prefix, 'layers'),
                                n_steps=nsteps,
                                truncate_gradient=options['truncate_grad'])
    # return all hidden states h(t)
    return rval[0]

def param_init_lstm_pskip(options, params, prefix='lstm_pskip', hier_level=False):
    """
    Init the LSTM-pskip parameter
    """
    # same as vanilla lstm layer
    params = param_init_lstm(options, params, prefix=prefix, hier_level=hier_level)
    # weight for skip connection
    params[_p(prefix, 'W_skip')] = numpy.array([numpy.random.random_sample()]).astype('float32')[0]
    # random value in (0,1)

    return params

def lstm_pskip_layer(tparams, state_below, options, prefix='lstm_pskip', mask=None):
    '''
    lstm layer with soft parametric weighted skip connections
    '''
    nsteps = state_below.shape[0]
    n_skip = options['skip_steps']
    if state_below.ndim == 3:
        n_samples = state_below.shape[1]
    else:
        n_samples = 1

    assert mask is not None

    def _slice(_x, n, dim):
        if _x.ndim == 3:
            return _x[:, :, n * dim:(n + 1) * dim]
        return _x[:, n * dim:(n + 1) * dim]

    def _lstm_unit(m_, x_, h_, c_, h_skip, hcnt):
        '''
        lstm_soft_pskip unit at each time step
        :param m_: mask
        :param x_: x(t) input
        :param h_: h(t-1) recurrent hidden state
        :param c_: c(t-1) cell state
        :param h_skip: h(t-n_skip) for skip connection
        :param hcnt: mark current time stamp (to determine whether skip connection exists)
        :return: h(t), c(t), h_skip, hcnt
        '''
        skip_flag = tensor.eq(hcnt % n_skip, 0)
        preact = tensor.dot(h_, tparams[_p(prefix, 'U')])
        preact += x_

        # gates
        i = tensor.nnet.sigmoid(_slice(preact, 0, options['dim_proj']))
        f = tensor.nnet.sigmoid(_slice(preact, 1, options['dim_proj']))
        o = tensor.nnet.sigmoid(_slice(preact, 2, options['dim_proj']))
        c = tensor.tanh(_slice(preact, 3, options['dim_proj']))

        # cell state
        c = f * c_ + i * c
        c = m_[:, None] * c + (1. - m_)[:, None] * c_
        # new hidden stae
        h = o * tensor.tanh(c) + h_skip * skip_flag * tparams[_p(prefix, 'W_skip')]
        h = m_[:, None] * h + (1. - m_)[:, None] * h_
        # update h_skip
        h_skip = h_skip * (1-skip_flag) + h * skip_flag
        hcnt += 1 # bug fixed T^T

        return h, c, h_skip, hcnt

    state_below = (tensor.dot(state_below, tparams[_p(prefix, 'W')]) +
                   tparams[_p(prefix, 'b')])

    dim_proj = options['dim_proj']
    h = tensor.alloc(numpy_floatX(0.), n_samples, dim_proj)
    c = tensor.alloc(numpy_floatX(0.), n_samples, dim_proj)
    h_skip = tensor.alloc(numpy_floatX(0.), n_samples, dim_proj)
    hcnt = tensor.zeros_like(theano.shared(10.).astype('float32'))
    rval, updates = theano.scan(_lstm_unit,
                                sequences=[mask, state_below],
                                outputs_info=[h, c, h_skip, hcnt],
                                name=_p(prefix, 'layers'),
                                n_steps=nsteps,
                                truncate_gradient=options['truncate_grad'])
    # return all hidden states h(t)
    return rval[0]

# ===========================
# residual (RRN) unit
# ===========================

def param_init_residual(options, params, prefix='residual'):
    """
    Init the residual_network parameter:
    """
    # weight for input x
    depth = options['unit_depth']
    Wx = dict()
    for idx in xrange(depth):
        Wx[idx] = norm_weight(options['dim_word'], options['dim_proj'])
    W = numpy.concatenate([ww for kk, ww in Wx.iteritems()], axis=1)
    params[_p(prefix, 'W')] = W
    b = numpy.zeros((depth * options['dim_proj'],))
    params[_p(prefix, 'b')] = b.astype(config.floatX)

    # weight for identity connection
    '''
    w_res = numpy.array([numpy.random.random_sample()]).astype('float32')[0]
    params[_p(prefix, 'w_res')] = w_res.astype(config.floatX)
    '''
    # weight for inter-states
    for idx in xrange(depth):
        U = ortho_weight(options['dim_proj'])
        params[_p(prefix, 'U'+str(idx+1))] = U

    return params

def residual_layer(tparams, state_below, options, prefix='residual', mask=None):
    '''
    vanilla residual layer (recurrent depth adjustable)
    '''
    # here state_below in x_emb
    nsteps = state_below.shape[0]
    depth = options['unit_depth']
    if state_below.ndim == 3:
        n_samples = state_below.shape[1]
    else:
        n_samples = 1

    assert mask is not None

    def _slice(_x, n, dim):
        if _x.ndim == 3:
            return _x[:, :, n * dim:(n + 1) * dim]
        return _x[:, n * dim:(n + 1) * dim]

    def _resblock(m_, x_, h_):
        y = h_
        for idx in xrange(depth):
            hy = tensor.dot(y, tparams[_p(prefix, 'U'+str(idx+1))])
            # y(i) = sigmoid(Wx(t)+b + Uy(i-1))
            y = tensor.nnet.sigmoid(_slice(x_, idx, options['dim_proj']) + hy)
        h = tensor.tanh(h_ + y)
        h = m_[:, None] * h + (1. - m_)[:, None] * h_
        return h

    # state_below = W*x(t)+b (for all inter_state y)
    state_below = (tensor.dot(state_below, tparams[_p(prefix, 'W')]) +
                   tparams[_p(prefix, 'b')])

    h = tensor.alloc(numpy_floatX(0.), n_samples, options['dim_proj'])
    rval, updates = theano.scan(_resblock,
                                sequences=[mask, state_below],
                                outputs_info=[h],
                                name=_p(prefix, 'layers'),
                                n_steps=nsteps,
                                truncate_gradient=options['truncate_grad'])
    return rval # bug fixed: not rval[0], attention here

# ===========================
# simple rnn
# ===========================

def param_init_rnn(options, params, prefix='rnn', hier_level=False):
    '''
    Initialize parameters for simple rnn unit
    Support hierarchical architecture
    '''
    if hier_level:
        W = ortho_weight(options['dim_proj'])
    else:
        W = norm_weight(options['dim_word'], options['dim_proj'])
    params[_p(prefix, 'W')] = W
    U = ortho_weight(options['dim_proj'])
    params[_p(prefix, 'U')] = U
    b = numpy.zeros((options['dim_proj']))
    params[_p(prefix, 'b')] = b.astype(config.floatX)

    return params

def rnn_layer(tparams, state_below, options, prefix='rnn', mask=None):
    nsteps = state_below.shape[0]
    dim_proj = options['dim_proj']
    if state_below.ndim == 3:
        n_samples = state_below.shape[1]
    else:
        n_samples = 1

    assert mask is not None

    # input: mask, x(t), h(t-1)
    def _rnn_unit(m_, x_, h_):
        h = tensor.tanh(tensor.dot(x_, tparams[_p(prefix, 'W')]) +
                        tensor.dot(h_, tparams[_p(prefix, 'U')]) +
                        tparams[_p(prefix, 'b')])
        h = m_[:, None] * h + (1.-m_)[:, None] * h_ # mask
        return h

    h = tensor.alloc(numpy_floatX(0.), n_samples, dim_proj)
    rval, updates = theano.scan(_rnn_unit,
                                sequences=[mask, state_below],
                                outputs_info=[h],
                                name=_p(prefix, 'layers'),
                                n_steps=nsteps,
                                truncate_gradient=options['truncate_grad'])
    return rval
