"""
Model for sequence classification tasks
supported model structures (refer to Layers.py for unit details):
    > simple-RNN
    > LSTM: standard-LSTM, hierarchical-LSTM
    > Recurrent Residual Network (RRN)
    > Skip-Connected LSTM (SC-LSTM):
        - with identity connections (SC-LSTM-I)
        - with parametric skip connections (SC-LSTM-P)
    > Hybrid Residual LSTM (HRL)
supported features:
    > dropout on non-recurrent layers (input, output)
    > gradient clipping
    > L2-regularization
"""
__author__ = 'v-yirwan'

import sys
import time

import numpy
import cPickle as pkl
import theano
import theano.tensor as tensor
from theano.sandbox.rng_mrg import MRG_RandomStreams as RandomStreams

from Layers import get_layer
from Data import *
from Util import *

# Set the random number generators' seeds for consistency
SEED = 123
numpy.random.seed(SEED)

def _p(pp, name):
    return '%s_%s' % (pp, name)

def init_params(options):
    """
    Global (not LSTM) parameter. For the embedding and the classifier.
    """
    params = OrderedDict()
    # embedding
    if options['dataset'] != 'mnist':
        randn = rand_weight(options['n_words'], options['dim_word'])
        params['Wemb'] = randn.astype(config.floatX)

    # encoder layer
    if options['hierarchical']:
         # word level LSTM layer
        params = get_layer(options['encoder'])[0](options, params,
                                                  prefix=options['encoder']+'_word')
        # sentence/phrase level LSTM layer
        params = get_layer(options['encoder'])[0](options, params,
                                                  prefix=options['encoder']+'_sen',
                                                  hier_level=True) # bug fixed: dim-matching
    elif options['hybrid']:
        params = get_layer(options['encoder'])[0](options, params,
                                                  prefix=options['encoder'])
        params = get_layer(options['encoder2'])[0](options, params,
                                                   prefix=options['encoder2'])
    else:
        params = get_layer(options['encoder'])[0](options, params,
                                                  prefix=options['encoder'])

    # classifier
    params['U'] = 0.01 * numpy.random.randn(options['dim_proj'],
                                            options['ydim']).astype(config.floatX)
    params['b'] = numpy.zeros((options['ydim'],)).astype(config.floatX)

    return params

def load_params(path, params):
    pp = numpy.load(path)
    for kk, vv in params.items():
        if kk not in pp:
            raise Warning('%s is not in the archive' % kk)
        params[kk] = pp[kk]

    return params

def init_tparams(params):
    tparams = OrderedDict()
    for kk, pp in params.items():
        tparams[kk] = theano.shared(params[kk], name=kk)
    return tparams

def encoder_word_layer(tparams, state_below, options, mask=None):
    '''
    word(bottom)-level encoder for hierarchical architecture
    '''
    def _encode(x_sub, mask_sub, proj_sub):
        n_timesteps = x_sub.shape[0]
        n_samples = x_sub.shape[1]
        emb_sub = tparams['Wemb'][x_sub.flatten()].reshape([n_timesteps, n_samples, options['dim_word']])
        proj_sub = get_layer(options['encoder'])[1](tparams, emb_sub, options,
                                                    prefix=options['encoder']+'_word',
                                                    mask=mask_sub)
        return proj_sub[-1]
    proj_sub = tensor.alloc(numpy_floatX(0.), state_below.shape[2], options['dim_proj'])
    rval, update = theano.scan(_encode,
                               sequences=[state_below, mask],
                               outputs_info=[proj_sub],
                               name='word_encoder_layer',
                               n_steps=state_below.shape[0])
    return rval

def build_model(tparams, options):
    trng = RandomStreams(SEED)

    # Used for dropout.
    use_noise = theano.shared(numpy_floatX(0.))

    if options['dataset'] == 'mnist':
        print 'Using mnist dataset with single number input'
        x = tensor.matrix('x', dtype='float32')
    else:
        print 'Using text dataset with embedding input'
        x = tensor.matrix('x', dtype='int64')
    mask = tensor.matrix('mask', dtype=config.floatX)
    y = tensor.vector('y', dtype='int64')

    n_timesteps = x.shape[0]
    n_samples = x.shape[1]

    # input word embedding
    if options['dataset'] == 'mnist':
        emb = x.reshape([n_timesteps, n_samples, options['dim_word']])
    else:
        emb = tparams['Wemb'][x.flatten()].reshape([n_timesteps, n_samples, options['dim_word']])

    # dropout on embedding
    if options['dropout_input'] > 0:
        print 'Applying drop-out on input embedding (dropout_input:', options['dropout_input'], ')'
        emb = dropout_layer(emb, options['dropout_input'], use_noise, trng)

    # encoder information
    print 'Using', options['encoder'], 'unit'
    if 'skip' in options['encoder']:
        print 'Skip connection length', options['skip_steps']
    if options['truncate_grad'] is not None and options['truncate_grad'] > 0:
        print 'Using gradient truncation to', options['truncate_grad'], 'steps'
    else:
        options['truncate_grad'] = -1

    # encoding layer
    proj = get_layer(options['encoder'])[1](tparams, emb, options,
                                            prefix=options['encoder'],
                                            mask=mask)

    # pooling
    if options['mean_pooling']:
        print 'Using mean_pooling'
        proj = (proj * mask[:, :, None]).sum(axis=0) # mean pooling
        proj = proj / mask.sum(axis=0)[:, None]
    else:
        print 'Using last hidden state'
        proj = proj[-1] # last hidden state

    sys.stdout.flush()

    # dropout on hidden states
    if options['dropout_output'] > 0:
        print 'Applying drop-out on hidden states (dropout_output:', options['dropout_output'], ")"
        proj = dropout_layer(proj, options['dropout_output'], use_noise, trng)

    pred = tensor.nnet.softmax(tensor.dot(proj, tparams['U']) + tparams['b'])

    # for training
    f_pred_prob = theano.function([x, mask], pred, name='f_pred_prob')
    f_pred = theano.function([x, mask], pred.argmax(axis=1), name='f_pred') # sample by argmax

    off = 1e-8
    if pred.dtype == 'float16':
        off = 1e-6
    cost = -tensor.log(pred[tensor.arange(n_samples), y] + off).mean()

    return use_noise, x, mask, y, f_pred_prob, f_pred, cost

def build_model_hybrid(tparams, options):
    trng = RandomStreams(SEED)

    # Used for dropout.
    use_noise = theano.shared(numpy_floatX(0.))

    if options['dataset'] == 'mnist':
        print 'Using mnist dataset with single number input'
        x = tensor.matrix('x', dtype='float32')
    else:
        print 'Using text dataset with embedding input'
        x = tensor.matrix('x', dtype='int64')
    mask = tensor.matrix('mask', dtype=config.floatX)
    y = tensor.vector('y', dtype='int64')

    n_timesteps = x.shape[0]
    n_samples = x.shape[1]

    # input word embedding
    if options['dataset'] == 'mnist':
        emb = x.reshape([n_timesteps, n_samples, options['dim_word']])
    else:
        emb = tparams['Wemb'][x.flatten()].reshape([n_timesteps, n_samples, options['dim_word']])

    # dropout on embedding
    if options['dropout_input'] > 0:
        print 'Applying drop-out on input embedding (dropout_input:', options['dropout_input'], ')'
        emb = dropout_layer(emb, options['dropout_input'], use_noise, trng)

    # encoder information
    if 'skip' in options['encoder']:
        print options['encoder'], 'unit: skip connection length', options['skip_steps']
    if 'skip' in options['encoder2']:
        print options['encoder2'], 'unit: skip connection length', options['skip_steps2']
    if options['truncate_grad'] is not None and options['truncate_grad'] > 0:
        print 'Using gradient truncation to', options['truncate_grad'], 'steps'
    else:
        options['truncate_grad'] = -1

    # encoding layer
    proj1 = get_layer(options['encoder'])[1](tparams, emb, options,
                                             prefix=options['encoder'],
                                             mask=mask)
    proj2 = get_layer(options['encoder2'])[1](tparams, emb, options,
                                              prefix=options['encoder2'],
                                              mask=mask)
    # pooling
    if options['mean_pooling']:
        print 'Using mean_pooling'
        proj1= (proj1 * mask[:, :, None]).sum(axis=0) # mean pooling
        proj1 = proj1 / mask.sum(axis=0)[:, None]
        proj2 = (proj2 * mask[:, :, None]).sum(axis=0)
        proj2 = proj2 / mask.sum(axis=0)[:, None]
    else:
        print 'Using last hidden state'
        proj1 = proj1[-1] # last hidden state
        proj2 = proj2[-1]

    # integrate
    proj = 0.5 * (proj1 + proj2)

    sys.stdout.flush()

    # dropout on hidden states
    if options['dropout_output'] > 0:
        print 'Applying drop-out on hidden states (dropout_output:', options['dropout_output'], ")"
        proj = dropout_layer(proj, options['dropout_output'], use_noise, trng)

    pred = tensor.nnet.softmax(tensor.dot(proj, tparams['U']) + tparams['b'])

    # for training
    f_pred_prob = theano.function([x, mask], pred, name='f_pred_prob')
    f_pred = theano.function([x, mask], pred.argmax(axis=1), name='f_pred') # sample by argmax

    off = 1e-8
    if pred.dtype == 'float16':
        off = 1e-6
    cost = -tensor.log(pred[tensor.arange(n_samples), y] + off).mean()

    return use_noise, x, mask, y, f_pred_prob, f_pred, cost

def build_model_hier(tparams, options):
    '''
    building model for hierarchical architecture
    support single-layer lstm unit now
    '''
    assert options['encoder']=='lstm', 'Use LSTM Unit for hierarchical mode'
    trng = RandomStreams(SEED)
    # Used for dropout.
    use_noise = theano.shared(numpy_floatX(0.))

    # word-level input
    x = tensor.tensor3('x', dtype='int64') # full batch input: n_batch * step_len * n_samples
    mask = tensor.tensor3('mask', dtype=config.floatX) # mask: n_batch * step_len * n_samples
    y = tensor.vector('y', dtype='int64')

    # word-level output hidden states, higher-level input
    # x_proj = tensor.alloc(numpy_floatX(0.), options['n_batch'], x.shape[2], options['dim_proj'])
    mask_proj = tensor.matrix('mask_proj', dtype=config.floatX) # mask for higher-level layer: n_batch * n_samples

    # word-level encoding
    print 'Using', options['encoder'], 'unit for word-level encoder'
    x_proj = encoder_word_layer(tparams, x, options, mask=mask)

    # sentence-level encoding
    print 'Using', options['encoder'], 'unit for sentence-level encoder'
    proj = get_layer(options['encoder'])[1](tparams, x_proj, options,
                                            prefix=options['encoder']+'_sen',
                                            mask=mask_proj)

    if options['mean_pooling']:
        print 'Using mean_pooling'
        proj = (proj * mask[:, :, None]).sum(axis=0) # mean pooling
        proj = proj / mask.sum(axis=0)[:, None]
    else:
        print 'Using last hidden state'
        proj = proj[-1] # last hidden state

    proj_test = proj

    if options['dropout_output'] is not None and options['dropout_output'] > 0:
        print 'Applying drop out (dropout_output: ', options['dropout_output'],')...',
        proj_test *= (1 - options['dropout_output'])
        proj = dropout_layer(proj, options['dropout_output'], use_noise, trng) # fixme: check dropout layer
        print 'Done'

    pred = tensor.nnet.softmax(tensor.dot(proj, tparams['U']) + tparams['b'])

    f_pred_prob = theano.function([x, mask, mask_proj], pred, name='f_pred_prob')
    f_pred = theano.function([x, mask, mask_proj], pred.argmax(axis=1), name='f_pred') # sample by argmax

    off = 1e-8
    if pred.dtype == 'float16':
        off = 1e-6
    cost = -tensor.log(pred[tensor.arange(x.shape[2]), y] + off).mean()

    return use_noise, x, mask, mask_proj, y, f_pred_prob, f_pred, cost

class Model:
    def __init__(self,
                 dim_word=128, # word embeding dimension
                 dim_proj=512,  # LSTM number of hidden units
                 patience=10,  # Number of epoch to wait before early stop if no progress
                 max_epochs=5000,  # The maximum number of epoch to run
                 decay_c=0.,  # Weight decay (for L2-regularization)
                 clip_c=-1., # gradient clipping threshold
                 lrate=1.,  # Learning rate for sgd (not used for adadelta and rmsprop)
                 n_words=10000,  # Vocabulary size
                 optimizer='adadelta',
                 encoder='lstm', # name of encoder unit, refer to 'layers'
                 encoder2=None, # only used in hybrid mode
                 hierarchical=False, # whether use hierarchical structure
                 hier_len=None, # length of bottom (word-level) encoder
                 hybrid=False, # whether use hybrid model
                 mean_pooling=True, # use last hidden state if false
                 unit_depth=2, # recurrent depth of residual unit
                 skip_steps=5, # skip connection length (h(t) -> h(t+skip_steps))
                 skip_steps2=-1, # only used in hybrid mode
                 truncate_grad=-1, # e number of steps to use in truncated BPTT, set to -1 if not to apply
                 saveto='model.npz',  # The best model will be saved there
                 dispFreq=50,  # Display the training progress after this number of updates
                 validFreq=300,  # Compute the validation error after this number of updates
                 newDumpFreq=5000, # Dump model into a new file after this number of updates
                 maxlen=100,  # Sequence longer then this get ignored
                 batch_size=64,  # The batch size during training.
                 batch_len_threshold=None, # use dynamic batch size if sequence lengths exceed this threshold
                 valid_batch_size=64,  # The batch size used for validation/test set.
                 dataset='text', # dataset dype
                 corpus='imdb.pkl', # path to load training data
                 start_iter=0,
                 start_epoch=0,
                 noise_std=0.,
                 dropout_output=None,  # Dropout on output hidden states (before softmax layer)
                 dropout_input=None, # Dropout on input embeddings
                 reload_options=None, # Path to a saved model options we want to start from
                 reload_model=None,  # Path to a saved model we want to start from.
                 embedding=None, # Path to the word embedding file (otherwise randomized)
                 test_size=-1,  # If >0, we keep only this number of test example.
                 monitor_grad=False, # Print gradient norm to log file at each iteration if set True
                 logFile='log.txt' # Path to log file
                 ):

        # Model options
        self.model_options = locals().copy()
        self.model_options['self'] = None

        # log files
        self.F_log = open(logFile, "a")
        if 'pskip' in self.model_options['encoder']:
            self.F_wskip = open(logFile.replace('log', 'wskip'), "a") # log file to monitor parametric skip weight
        if monitor_grad:
            self.F_grad = open(logFile.replace('log', 'log_grad'), "a") # log file to monitor gradients

        if start_iter == 0:
            self.F_log.write("model options:\n")
            for kk, vv in self.model_options.iteritems():
                self.F_log.write("\t"+kk+":\t"+str(vv)+"\n")
            self.F_log.write("-----------------------------------------\n")

        pkl.dump(self.model_options, open('%s.pkl' % saveto, 'wb'))

        print 'Loading data...',
        if dataset == 'mnist':
            self.trainSet, self.validSet, self.testSet = load_mnist(path=corpus,
                                                                    fixed_permute=True,
                                                                    rand_permute=False)
        else:
            self.trainSet, self.validSet, self.testSet = load_data(path=corpus,
                                                                   n_words=n_words,
                                                                   maxlen=maxlen,
                                                                   sort_by_len=True,
                                                                   fixed_valid=True)
        print 'Done! '
        print 'Training', len(self.trainSet[0]), 'Valid', len(self.validSet[0]), 'Test', len(self.testSet[0])
        sys.stdout.flush()

        if test_size > 0:
            test_size = min(test_size, len(self.testSet[0]))
            idx = numpy.arange(len(self.testSet[0]))
            numpy.random.shuffle(idx)
            idx = idx[:test_size]
            self.testSet = ([self.testSet[0][n] for n in idx], [self.testSet[1][n] for n in idx])

        # number of classes
        ydim = numpy.max(self.trainSet[1]) + 1
        self.model_options['ydim'] = ydim

        print 'Initializing model parameters...',
        params = init_params(self.model_options)
        print 'Done'
        print 'Model size:', self.model_options['dim_word'], '*', self.model_options['dim_proj']
        sys.stdout.flush()

        # load pre-trained word embedding
        if embedding is not None and os.path.exists(embedding):
            Wemb = numpy.array(numpy.load(open(embedding, "rb")))
            if Wemb.shape[0] == self.model_options['n_words'] and \
                            Wemb.shape[1] == self.model_options['dim_word']:
                print 'Using pre-trained word embedding'
                params['Wemb'] = Wemb.astype(numpy.float32) # bug fixed
                print 'vocab size', params['Wemb'].shape[0], ', dim', params['Wemb'].shape[1]

        # reload options
        if reload_options is not None and os.path.exists(reload_options):
            print "Reloading model options...",
            with open(reload_options, 'rb') as f:
                self.model_options = pkl.load(f)
            print "Done"

        # reload parameters
        self.start_iter = 0
        self.start_epoch = 0
        self.history_errs = []
        if reload_model is not None and os.path.exists(reload_model): # bug fixed
            print 'Reloading model parameters...',
            load_params(reload_model, params)
            self.start_iter = start_iter
            self.start_epoch = start_epoch
            self.history_errs = list(numpy.load(self.model_options['reload_model'])['history_errs'])
            print 'Done'
        sys.stdout.flush()

        self.tparams = init_tparams(params)

        # build model
        mask_proj = None
        if self.model_options['hierarchical']:
            # hierarchical structure
            assert self.model_options['hier_len'] > 0, 'set hier_len>0 in hierarchical mode'
            print 'Using hierarchical structure, hier_step', self.model_options['hier_len']
            self.use_noise, x, mask, mask_proj, y, \
            self.f_pred_prob, self.f_pred, cost = \
                build_model_hier(self.tparams, self.model_options)
            inps = [x, mask, mask_proj, y]
        elif self.model_options['hybrid']:
            # hybrid structure
            print 'Using hybrid structure with', \
                self.model_options['encoder'], 'unit and', \
                self.model_options['encoder2'], 'unit'
            self.use_noise, x, mask, y, \
            self.f_pred_prob, self.f_pred, cost = \
                build_model_hybrid(self.tparams, self.model_options)
            inps = [x, mask, y]
        else:
            # vanilla structure
            print 'Using vanilla structure'
            self.use_noise, x, mask, y, \
            self.f_pred_prob, self.f_pred, cost = \
                build_model(self.tparams, self.model_options)
            inps = [x, mask, y]
        sys.stdout.flush()

        if decay_c > 0.:
            print 'Applying L2 regularization...',
            cost = l2_regularization(self.tparams, cost, decay_c)
            print 'Done'

        print 'Building f_cost...',
        self.f_cost = theano.function(inps, cost, name='f_cost')
        print 'Done'

        print 'Computing gradient...',
        grads = tensor.grad(cost, wrt=list(self.tparams.values()))
        self.f_grad = theano.function(inps, grads, name='f_grad')
        print 'Done'

        # monitor gradient
        if monitor_grad:
            print 'Computing gradient norm ...',
            g2 = 0.
            for g in grads:
                g2 += (g**2).sum()
            self.f_norm = theano.function(inps, tensor.sqrt(g2), name='f_norm')
            g_norm = []
            for g in grads:
                g2 = (g**2).sum()
                g_norm.append(tensor.sqrt(g2))
            self.f_norm_all = theano.function(inps, g_norm, name='f_norm_all')
            print 'Done'

        # apply gradient clipping here
        if clip_c > 0.:
            print 'Applying gradient clipping (clip_c:'+str(clip_c)+')...',
            grads = grad_clipping(grads, clip_c)
            print 'Done'

        print 'Buidling optimizers...',
        lr = tensor.scalar(name='lr')
        self.f_grad_shared, self.f_update = eval(optimizer)(lr, self.tparams, grads,
                                                            x, mask, y, cost,
                                                            mask_hier=mask_proj)
        print 'Done'
        sys.stdout.flush()

        self.kf_train = None
        if batch_len_threshold:
            # dynamic batch size with fixed bucket settings
            print 'Using dynamic batch size'
            self.kf_valid = get_minibatches_idx_bucket(self.validSet[0], valid_batch_size)
            self.kf_test = get_minibatches_idx_bucket(self.testSet[0], valid_batch_size)
        else:
            self.kf_valid = get_minibatches_idx(len(self.validSet[0]), valid_batch_size)
            self.kf_test = get_minibatches_idx(len(self.testSet[0]), valid_batch_size)

    def monitor_grad(self, x, y, mask, mask_proj, cost, uidx):
        self.F_grad.write('Iteration: '+str(uidx)+'\nLoss: '+str(cost)+'\n')
        self.F_grad.write('Batch length: '+str(x.shape[0])+'\n')
        self.F_grad.write('Parametric weight: ' +
                          str(self.tparams[_p(self.model_options['encoder'],'w_res')].get_value())+'\n')
        p2 = 0
        for kk, vv in self.tparams.iteritems():
            p2 += (self.tparams[kk].get_value()**2).sum()
            p2 = numpy.sqrt(p2)
            self.F_grad.write('Weight norm: '+str(p2)+'\n')

            if self.model_options['hierarchical']:
                g2 = self.f_norm(x, mask, mask_proj, y)
                g_norm = self.f_norm_all(x, mask, mask_proj, y)
            else:
                g2 = self.f_norm(x, mask, mask_proj, y)
                g_norm = self.f_norm_all(x, mask, y)
            self.F_grad.write('Gradient norm: '+str(g2)+'\n')

            g_name = [kk for kk, vv in self.tparams.iteritems()]
            self.F_grad.write('---------------------------------------\nWeight norms:\n')
            for kk, vv in self.tparams.iteritems():
                self.F_grad.write(kk+':\t'+
                                  str(numpy.sqrt((self.tparams[kk].get_value()**2).sum()))+'\n')

            self.F_grad.write('---------------------------------------\nGradient norms:\n')
            for i in range(len(g_norm)):
                self.F_grad.write(g_name[i]+':\t'+str(g_norm[i])+'\n')
            self.F_grad.write('\n')

            if g2 >= 50000:
                saveFile = self.model_options['saveto'] + "_iter_" + str(uidx) + "_loss_" + str(cost)
                params = unzip(self.tparams)
                numpy.savez(saveFile, history_errs=self.history_errs, **params)

    def display(self, cost_accu, cost, eidx, uidx, batch_len, batch_size):
        print 'Epoch ', eidx, '\tUpdate ', uidx, '\tCost ', cost_accu/float(self.model_options['dispFreq']),
        print '\tLength', batch_len, '\tSize ', batch_size
        self.F_log.write('Epoch '+str(eidx)+'\tUpdate '+str(uidx)+
                         '\tCost '+str(cost_accu/float(self.model_options['dispFreq']))
                         +'\tLength '+str(batch_len)+'\tSize '+str(batch_size)+'\n')
        # keep track of parameter w_skip
        if 'pskip' in self.model_options['encoder']:
            w_skip = self.tparams[_p(self.model_options['encoder'], 'W_skip')].get_value()
            self.F_wskip.write('Epoch '+str(eidx)+'\tUpdate '+str(uidx)+'\tW_skip '+
                           str(w_skip)+'\tLoss '+str(cost)+'\n')

    def get_accu(self, data, iterator, hier_len=None):
        """
        Just compute the error
        modified to support hierarchical mode
        """
        valid_acc = 0
        for _, valid_index in iterator:
            if hier_len is not None:
                x, mask, mask_proj, y = prepare_data_hier([data[0][t] for t in valid_index],
                                                           numpy.array(data[1])[valid_index],
                                                           hier_len=hier_len)
                preds = self.f_pred(x, mask, mask_proj)
            else:
                x, mask, y = prepare_data([data[0][t] for t in valid_index],
                                          numpy.array(data[1])[valid_index],
                                          maxlen=None,
                                          dataset=self.model_options['dataset'])
                preds = self.f_pred(x, mask) # result obtained by argmax
            valid_acc += (preds == y).sum() # note that batch is sorted in hier-mode
        valid_acc = numpy_floatX(valid_acc) / numpy_floatX(len(data[0])) # accuracy
        return valid_acc

    def save_model(self, savefile, best_p=None):
        if best_p is not None: # save the best model so far
            params = best_p
        else:
            params = unzip(self.tparams)
        numpy.savez(savefile, history_errs=self.history_errs, **params)
        pkl.dump(self.model_options, open('%s.pkl' % self.model_options['saveto'], 'wb'))

    def valid(self):
        train_acc = self.get_accu(self.trainSet, self.kf_train,
                                  hier_len=self.model_options['hier_len'])
        valid_acc = self.get_accu(self.validSet, self.kf_valid,
                                  hier_len=self.model_options['hier_len'])
        test_acc = self.get_accu(self.testSet, self.kf_test,
                                 hier_len=self.model_options['hier_len'])
        return train_acc, valid_acc, test_acc

    def train_minibatch(self, x, y, n_samples):
        if self.model_options['hierarchical']:
            x, mask, mask_proj, y = prepare_data_hier(x, y,
                                                      self.model_options['hier_len'],
                                                      dataset=self.model_options['dataset'])
            n_samples += x.shape[2]
            cost = self.f_grad_shared(x, mask, mask_proj, y)
            self.f_update(self.model_options['lrate'])
        else:
            x, mask, y = prepare_data(x, y, dataset=self.model_options['dataset'])
            mask_proj = None
            n_samples += x.shape[1]
            cost = self.f_grad_shared(x, mask, y)
            self.f_update(self.model_options['lrate'])
        return cost, n_samples, x, y, mask, mask_proj

    def train(self):
        best_p = None
        bad_count = 0

        if self.model_options['validFreq'] == -1:
            self.model_options['validFreq'] = len(self.trainSet[0]) // self.model_options['batch_size']

        uidx = self.start_iter # iteration idx
        estop = False  # early stop
        start_time = time.time()
        cost_accu = 0
        patience = self.model_options['patience']

        for eidx in xrange(self.start_epoch, self.model_options['max_epochs']):
            print 'Start epoch', eidx
            epoch_start_time = time.time()
            n_samples = 0

            if self.model_options['batch_len_threshold'] > 0:
                self.kf_train = get_minibatches_idx_bucket(self.trainSet[0],
                                                           self.model_options['batch_size'],
                                                           shuffle=True)
            else:
                # Get new shuffled index for the training set.
                self.kf_train = get_minibatches_idx(len(self.trainSet[0]),
                                                    self.model_options['batch_size'],
                                                    shuffle=True)

            for _, train_index in self.kf_train:
                uidx += 1
                self.use_noise.set_value(1.) # training mode dropout

                # Select the random examples for this minibatch
                x = [self.trainSet[0][t] for t in train_index]
                y = [self.trainSet[1][t] for t in train_index]

                cost, n_samples, \
                    x, y, mask, mask_proj = self.train_minibatch(x, y, n_samples)
                cost_accu += cost

                if numpy.isnan(cost) or numpy.isinf(cost):
                    print 'NAN detected: ', cost
                    self.F_log.write('bad cost detected\n Epoch' + str(eidx) + '\tUpdate ' +
                                     str(uidx) + '\tlength ' + str(x.shape[0]) + '\n')
                    return 1., 1., 1.

                # monitor gradient: print gradient norm, batch length at each iteration
                if self.model_options['monitor_grad']:
                    self.monitor_grad(x, y, mask, mask_proj, cost, uidx)

                if numpy.mod(uidx, self.model_options['dispFreq']) == 0:
                    self.display(cost_accu, cost, eidx, uidx, x.shape[0], x.shape[1])
                    cost_accu = 0.

                if numpy.mod(uidx, self.model_options['validFreq']) == 0:
                    training_time = time.time()
                    print 'Validating...',
                    self.use_noise.set_value(0.) # test mode: rescale
                    train_acc, valid_acc, test_acc = self.valid()
                    valid_err = 1. - valid_acc
                    self.history_errs.append([valid_err, 1.-test_acc]) # error rate history
                    print 'Done'

                    if (best_p is None or valid_err <= numpy.array(self.history_errs)[:, 0].min()):
                        best_p = unzip(self.tparams)
                        bad_count = 0

                    validation_time = time.time()
                    start_time = start_time + (validation_time - training_time)
                    # validation time is not included
                    current_time = validation_time - start_time
                    print '--------------------------------'
                    print 'Accuracy: Train', train_acc, 'Valid', valid_acc, 'Test', test_acc
                    print 'Time used:', current_time
                    print '--------------------------------'
                    self.F_log.write("================================\nAccuracy:\n" +
                                     "Train " + str(train_acc) + "\tValid " + str(valid_acc) +
                                     "\tTest " + str(test_acc) + "\nTime used " +
                                     str(current_time) + "\n================================\n")
                    sys.stdout.flush()

                    print 'Saving...',
                    savefile = self.model_options['saveto']
                    if numpy.mod(uidx, self.model_options['newDumpFreq']) == 0:
                        savefile = savefile.replace('.npz', '_e' + str(eidx) + '_i' + str(uidx) +
                                                    '_valid_' + str(valid_err) + '.npz')
                    self.save_model(savefile, best_p)
                    print 'Done'

                    if (len(self.history_errs) > patience and
                            valid_err >= numpy.array(self.history_errs)[:-patience, 0].min()):
                        bad_count += 1
                        if bad_count > patience:
                            print('Early Stop!')
                            self.F_log.write('####################\nEarly Stop!\n####################\n')
                            estop = True
                            break

            epoch_end_time = time.time()
            print "Epoch", eidx, "completed, Seen", n_samples, "samples, Time", (epoch_end_time - epoch_start_time)
            sys.stdout.flush()
            self.F_log.write("--------------------------------------\nEpoch " + str(eidx) +
                             " completed, Seen " + str(n_samples) + " samples, Time " +
                             str(epoch_end_time-epoch_start_time) + '\n' +
                             '--------------------------------------\n')
            if estop:
                break

        end_time = time.time()

        if best_p is not None:
            zipp(best_p, self.tparams)
        else:
            best_p = unzip(self.tparams)
        self.use_noise.set_value(0.) # test mode: rescale
        if self.model_options['batch_len_threshold']:
            self.kf_train = get_minibatches_idx_bucket(self.trainSet[0],
                                                       self.model_options['batch_size']) # not shuffled
        else:
            self.kf_train = get_minibatches_idx(len(self.trainSet[0]),
                                                self.model_options['batch_size'])
        train_acc, valid_acc, test_acc = self.valid()

        print '--------------------------------'
        print 'Accuracy: Train', train_acc, 'Valid', valid_acc, 'Test', test_acc
        print 'Time used:', end_time - start_time
        print '--------------------------------'
        self.F_log.write("================================\nAccuracy:\n"
                         "Train " + str(train_acc) + "\tValid " + str(valid_acc) + "\tTest " +
                         str(test_acc) + "\n================================\n")

        savefile = self.model_options['saveto'].replace('.npz', '_e' + str(eidx) + '_i' + str(uidx) +
                                                        '_acc_' + str(valid_acc) + '.npz')
        numpy.savez(savefile, train_err=train_acc,
                    valid_err=valid_acc, test_err=test_acc,
                    history_errs=self.history_errs, **best_p)

        print 'The code run for %d epochs, with %f sec/epochs' % \
              ((eidx + 1), (end_time - start_time) / (1. * (eidx + 1)))
        print 'Training took %.1fs' % (end_time - start_time)
        self.F_log.write('The code run for %d epochs, with %f sec/epochs' %
                         ((eidx + 1), (end_time - start_time) / (1. * (eidx + 1))) +
                         '\nTraining took %.1fs\n' % (end_time - start_time))
        self.F_log.close()
        return train_acc, valid_acc, test_acc


if __name__ == '__main__':
    pass




